@@ -20,16 +20,18 @@ from ape.exceptions import (
     OutOfGasError,
     ProviderError,
     SubprocessError,
-    TransactionError,
     VirtualMachineError,
 )
 from ape.logging import logger
 from ape.types import AddressType, SnapshotID
-from ape.utils import cached_property, gas_estimation_error_message
+from ape.utils import cached_property
 from ape_test import Config as TestConfig
-from evm_trace import TraceFrame
+from eth_utils import is_0x_prefixed, to_hex
+from evm_trace import CallTreeNode, CallType, TraceFrame, get_calltree_from_geth_trace
+from hexbytes import HexBytes
 from web3 import HTTPProvider, Web3
 from web3.gas_strategies.rpc import rpc_gas_price_strategy
+from web3.middleware import geth_poa_middleware
 
 from .exceptions import HardhatNotInstalledError, HardhatProviderError, HardhatSubprocessError
 
@@ -57,6 +59,10 @@ module.exports = {{
 }};
 """
 HARDHAT_CONFIG_FILE_NAME = "hardhat.config.js"
+_NO_REASON_REVERT_MESSAGE = "Transaction reverted without a reason string"
+_REVERT_REASON_PREFIX = (
+    "Error: VM Exception while processing transaction: reverted with reason string "
+)
 
 
 class HardhatConfigJS:
@@ -64,8 +70,6 @@ class HardhatConfigJS:
     A class representing the actual ``hardhat.config.js`` file.
     """
 
-    FILE_NAME = "hardhat.config.js"
-
     def __init__(
         self,
         project_path: Path,
@@ -86,7 +90,7 @@ class HardhatConfigJS:
 
     @property
     def _path(self) -> Path:
-        return self._base_path / self.FILE_NAME
+        return self._base_path / HARDHAT_CONFIG_FILE_NAME
 
     def write_if_not_exists(self):
         if not self._path.is_file():
@@ -104,9 +108,11 @@ class HardhatNetworkConfig(PluginConfig):
     # Retry strategy configs, try increasing these if you're getting HardhatSubprocessError
     network_retries: List[float] = HARDHAT_START_NETWORK_RETRIES
     process_attempts: int = HARDHAT_START_PROCESS_ATTEMPTS
+    request_timeout: int = 30
+    fork_request_timeout: int = 300
 
     # For setting the values in --fork and --fork-block-number command arguments.
-    # Used only in HardhatMainnetForkProvider.
+    # Used only in HardhatForkProvider.
     # Mapping of ecosystem_name => network_name => HardhatForkConfig
     fork: Dict[str, Dict[str, HardhatForkConfig]] = {}
 
@@ -135,10 +141,14 @@ class HardhatProvider(SubprocessProvider, Web3Provider, TestProviderAPI):
     def process_name(self) -> str:
         return "Hardhat node"
 
+    @property
+    def timeout(self) -> int:
+        return self.config.request_timeout  # type: ignore
+
     @property
     def chain_id(self) -> int:
-        if hasattr(self._web3, "eth"):
-            return self._web3.eth.chain_id
+        if hasattr(self.web3, "eth"):
+            return self.web3.eth.chain_id
         else:
             return HARDHAT_CHAIN_ID
 
@@ -231,7 +241,7 @@ class HardhatProvider(SubprocessProvider, Web3Provider, TestProviderAPI):
         if not self.port:
             return
 
-        self._web3 = Web3(HTTPProvider(self.uri))
+        self._web3 = Web3(HTTPProvider(self.uri, request_kwargs={"timeout": self.timeout}))
         if not self._web3.isConnected():
             self._web3 = None
             return
@@ -262,7 +272,7 @@ class HardhatProvider(SubprocessProvider, Web3Provider, TestProviderAPI):
                     port = random.randint(EPHEMERAL_PORTS_START, EPHEMERAL_PORTS_END)
                     attempts += 1
                     if attempts == max_attempts:
-                        ports_str = ", ".join(self.attempted_ports)
+                        ports_str = ", ".join([str(p) for p in self.attempted_ports])
                         raise HardhatProviderError(
                             f"Unable to find an available port. Ports tried: {ports_str}"
                         )
@@ -289,10 +299,11 @@ class HardhatProvider(SubprocessProvider, Web3Provider, TestProviderAPI):
         ]
 
     def _make_request(self, rpc: str, args: list) -> Any:
-        return self._web3.manager.request_blocking(rpc, args)  # type: ignore
+        return self.web3.provider.make_request(rpc, args)  # type: ignore
 
     def set_block_gas_limit(self, gas_limit: int) -> bool:
-        return self._make_request("evm_setBlockGasLimit", [hex(gas_limit)])
+        result = self._make_request("evm_setBlockGasLimit", [hex(gas_limit)])
+        return result.get("result") is True
 
     def set_timestamp(self, new_timestamp: int):
         pending_timestamp = self.get_block("pending").timestamp
@@ -300,18 +311,20 @@ class HardhatProvider(SubprocessProvider, Web3Provider, TestProviderAPI):
         self._make_request("evm_increaseTime", [seconds_to_increase])
 
     def mine(self, num_blocks: int = 1):
-        for i in range(num_blocks):
-            self._make_request("evm_mine", [])
+        # NOTE: Request fails when given numbers with any left padded 0s.
+        num_blocks_arg = f"0x{HexBytes(num_blocks).hex().replace('0x', '').lstrip('0')}"
+        self._make_request("hardhat_mine", [num_blocks_arg])
 
     def snapshot(self) -> str:
         result = self._make_request("evm_snapshot", [])
-        return str(result)
+        return result["result"]
 
-    def revert(self, snapshot_id: SnapshotID):
-        if isinstance(snapshot_id, str) and snapshot_id.isnumeric():
-            snapshot_id = int(snapshot_id)  # type: ignore
+    def revert(self, snapshot_id: SnapshotID) -> bool:
+        if isinstance(snapshot_id, int):
+            snapshot_id = HexBytes(snapshot_id).hex()
 
-        return self._make_request("evm_revert", [snapshot_id])
+        result = self._make_request("evm_revert", [snapshot_id])
+        return result.get("result") is True
 
     def unlock_account(self, address: AddressType) -> bool:
         result = self._make_request("hardhat_impersonateAccount", [address])
@@ -319,26 +332,7 @@ class HardhatProvider(SubprocessProvider, Web3Provider, TestProviderAPI):
         if result:
             self.unlocked_accounts.append(address)
 
-        return result
-
-    def estimate_gas_cost(self, txn: TransactionAPI) -> int:
-        """
-        Generates and returns an estimate of how much gas is necessary
-        to allow the transaction to complete.
-        The transaction will not be added to the blockchain.
-        """
-        try:
-            return super().estimate_gas_cost(txn)
-        except ValueError as err:
-            tx_error = _get_vm_error(err)
-
-            # If this is the cause of a would-be revert,
-            # raise ContractLogicError so that we can confirm tx-reverts.
-            if isinstance(tx_error, ContractLogicError):
-                raise tx_error from err
-
-            message = gas_estimation_error_message(tx_error)
-            raise TransactionError(base_err=tx_error, message=message) from err
+        return result.get("result") is True
 
     def send_transaction(self, txn: TransactionAPI) -> ReceiptAPI:
         """
@@ -346,45 +340,94 @@ class HardhatProvider(SubprocessProvider, Web3Provider, TestProviderAPI):
         for signed transactions.
         """
 
-        sender = self.conversion_manager.convert(txn.sender, AddressType)
+        sender = txn.sender
+        if sender:
+            sender = self.conversion_manager.convert(txn.sender, AddressType)
+
         if sender in self.unlocked_accounts:
             # Allow for an unsigned transaction
+            txn = self.prepare_transaction(txn)
             txn_dict = txn.dict()
 
             try:
                 txn_hash = self._web3.eth.send_transaction(txn_dict)  # type: ignore
             except ValueError as err:
-                raise _get_vm_error(err) from err
+                raise self.get_virtual_machine_error(err) from err
 
             receipt = self.get_transaction(
                 txn_hash.hex(), required_confirmations=txn.required_confirmations or 0
             )
+            receipt.raise_for_status()
 
         else:
-            try:
-                receipt = super().send_transaction(txn)
-            except ValueError as err:
-                raise _get_vm_error(err) from err
+            receipt = super().send_transaction(txn)
 
-        receipt.raise_for_status()
         return receipt
 
     def get_transaction_trace(self, txn_hash: str) -> Iterator[TraceFrame]:
-        """
-        Provide a detailed description of opcodes.
+        result = self._make_request("debug_traceTransaction", [txn_hash])
+        frames = result.get("structLogs", [])
+        for frame in frames:
+            yield TraceFrame(**frame)
 
-        Args:
-            txn_hash (str): The hash of a transaction to trace.
+    def get_call_tree(self, txn_hash: str) -> CallTreeNode:
+        receipt = self.get_transaction(txn_hash)
+        root_node_kwargs = {
+            "gas_cost": receipt.gas_used,
+            "gas_limit": receipt.gas_limit,
+            "address": receipt.receiver,
+            "calldata": receipt.data,
+            "value": receipt.value,
+            "call_type": CallType.CALL,
+            "failed": receipt.failed,
+        }
+        return get_calltree_from_geth_trace(receipt.trace, **root_node_kwargs)
 
-        Returns:
-            Iterator(TraceFrame): Transaction execution trace object.
-        """
-        logs = self._make_request("debug_traceTransaction", [txn_hash]).structLogs
-        for log in logs:
-            yield TraceFrame(**log)
+    def set_balance(self, account: AddressType, amount: Union[int, float, str, bytes]):
+        is_str = isinstance(amount, str)
+        is_hex = False if not is_str else is_0x_prefixed(str(amount))
+        is_key_word = is_str and len(str(amount).split(" ")) > 1
+        if is_key_word:
+            # This allows values such as "1000 ETH".
+            amount = self.conversion_manager.convert(amount, int)
+            is_str = False
+
+        amount_hex_str = str(amount)
+
+        # Convert to hex str
+        if is_str and not is_hex:
+            amount_hex_str = to_hex(int(amount))
+        elif isinstance(amount, int) or isinstance(amount, bytes):
+            amount_hex_str = to_hex(amount)
+
+        self._make_request("hardhat_setBalance", [account, amount_hex_str])
+
+    def get_virtual_machine_error(self, exception: Exception) -> VirtualMachineError:
+        if not len(exception.args):
+            return VirtualMachineError(base_err=exception)
+
+        err_data = exception.args[0]
+
+        message = err_data if isinstance(err_data, str) else str(err_data.get("message"))
+        if not message:
+            return VirtualMachineError(base_err=exception)
+        elif message.startswith("execution reverted: "):
+            message = message.replace("execution reverted: ", "")
+
+        if message.startswith(_REVERT_REASON_PREFIX):
+            message = message.replace(_REVERT_REASON_PREFIX, "").strip("'")
+            return ContractLogicError(revert_message=message)
+
+        elif _NO_REASON_REVERT_MESSAGE in message:
+            return ContractLogicError()
+
+        elif message == "Transaction ran out of gas":
+            return OutOfGasError()  # type: ignore
+
+        return VirtualMachineError(message=message)
 
 
-class HardhatMainnetForkProvider(HardhatProvider):
+class HardhatForkProvider(HardhatProvider):
     """
     A Hardhat provider that uses ``--fork``, like:
     ``npx hardhat node --fork <upstream-provider-url>``.
@@ -394,6 +437,18 @@ class HardhatMainnetForkProvider(HardhatProvider):
     to use as your archive node.
     """
 
+    @property
+    def fork_url(self) -> str:
+        return self._upstream_provider.connection_str  # type: ignore
+
+    @property
+    def fork_block_number(self) -> Optional[int]:
+        return self._fork_config.block_number
+
+    @property
+    def timeout(self) -> int:
+        return self.config.fork_request_timeout  # type: ignore
+
     @property
     def _upstream_network_name(self) -> str:
         return self.network.name.replace("-fork", "")
@@ -432,15 +487,19 @@ class HardhatMainnetForkProvider(HardhatProvider):
 
         # Verify that we're connected to a Hardhat node with mainnet-fork mode.
         self._upstream_provider.connect()
+        upstream_chain_id = self._upstream_provider.chain_id
         upstream_genesis_block_hash = self._upstream_provider.get_block(0).hash
         self._upstream_provider.disconnect()
+
+        # If upstream network is rinkeby, goerli, or kovan (PoA test-nets)
+        if upstream_chain_id in (4, 5, 42):
+            self._web3.middleware_onion.inject(geth_poa_middleware, layer=0)
+
         if self.get_block(0).hash != upstream_genesis_block_hash:
-            # self.disconnect()
             logger.warning(
                 "Upstream network has mismatching genesis block. "
                 "This could be an issue with hardhat."
             )
-            # raise HardhatProviderError(f"Upstream network is not {self._upstream_network_name}")
 
     def build_command(self) -> List[str]:
         if not isinstance(self._upstream_provider, UpstreamProvider):
@@ -448,48 +507,22 @@ class HardhatMainnetForkProvider(HardhatProvider):
                 f"Provider '{self._upstream_provider.name}' is not an upstream provider."
             )
 
-        fork_url = self._upstream_provider.connection_str
-        if not fork_url:
+        if not self.fork_url:
             raise HardhatProviderError("Upstream provider does not have a ``connection_str``.")
 
-        if fork_url.replace("localhost", "127.0.0.1") == self.uri:
+        if self.fork_url.replace("localhost", "127.0.0.1") == self.uri:
             raise HardhatProviderError(
                 "Invalid upstream-fork URL. Can't be same as local Hardhat node."
             )
 
         cmd = super().build_command()
-        cmd.extend(("--fork", fork_url))
-        fork_block_number = self._fork_config.block_number
-        if fork_block_number is not None:
-            cmd.extend(("--fork-block-number", str(fork_block_number)))
+        cmd.extend(("--fork", self.fork_url))
+        if self.fork_block_number is not None:
+            cmd.extend(("--fork-block-number", str(self.fork_block_number)))
 
         return cmd
 
-
-def _get_vm_error(web3_value_error: ValueError) -> TransactionError:
-    if not len(web3_value_error.args):
-        return VirtualMachineError(base_err=web3_value_error)
-
-    err_data = web3_value_error.args[0]
-    if not isinstance(err_data, dict):
-        return VirtualMachineError(base_err=web3_value_error)
-
-    message = str(err_data.get("message"))
-    if not message:
-        return VirtualMachineError(base_err=web3_value_error)
-
-    # Handle `ContactLogicError` similarly to other providers in `ape`.
-    # by stripping off the unnecessary prefix that hardhat has on reverts.
-    hardhat_prefix = (
-        "Error: VM Exception while processing transaction: reverted with reason string "
-    )
-    if message.startswith(hardhat_prefix):
-        message = message.replace(hardhat_prefix, "").strip("'")
-        return ContractLogicError(revert_message=message)
-    elif "Transaction reverted without a reason string" in message:
-        return ContractLogicError()
-
-    elif message == "Transaction ran out of gas":
-        return OutOfGasError()
-
-    return VirtualMachineError(message=message)
+    def reset_fork(self):
+        self._make_request(
+            "hardhat_reset", [{"jsonRpcUrl": self.fork_url, "blockNumber": self.fork_block_number}]
+        )
